package org.roboquant.brokers

import de.siegmar.fastcsv.reader.CsvReader
import org.roboquant.common.Config
import org.roboquant.common.Currency
import org.roboquant.common.Logging
import java.io.File
import java.io.InputStream
import java.io.InputStreamReader
import java.net.URL
import java.nio.file.Files
import java.nio.file.StandardCopyOption
import java.time.Instant
import java.time.LocalDate
import java.time.ZoneId
import java.time.ZonedDateTime
import java.time.format.DateTimeFormatter
import java.util.*
import java.util.logging.Logger
import java.util.zip.ZipInputStream
import kotlin.io.path.div
import kotlin.io.path.notExists

/**
 * Currency converter that uses the exchange reference rates as published by the ECB and that are freely available at the
 * [ECB website](https://www.ecb.europa.eu/stats/policy_and_exchange_rates/euro_reference_exchange_rates/html/index.en.html)
 *
 * It contains many daily conversion rates, and with this file loaded it is possible to trade in almost any currency
 * in the world. However, please note that:
 *
 * 1. The published rates are all in relationship to the Euro, so they only go back to the introduction date of
 * the Euro and don't cover earlier periods
 * 2. Cryptocurrencies are not included in this file
 *
 * # Example usage
 *
 *      val exchangeRates = ECBReferenceRates("data/RATES/rates.csv")
 *      val account = DefaultAccount(exchangeRates)
 *
 * @constructor Create a new  ECB exchange rates converter
 */
class ECBExchangeRates(url: String, compressed: Boolean = false, useCache: Boolean = false) : CurrencyConverter {

    private val converters: NavigableMap<Instant, FixedExchangeRates>
    private val logger: Logger = Logging.getLogger("ECBExchangeRates")

    val currencies: Collection<Currency>
        get() = converters.map { it.value.exchangeRates.keys }.flatten().toSet()

    val timeline: Collection<Instant>
        get() = converters.keys

    companion object {
        private const val DEFAULT_ECB_URL = "https://www.ecb.europa.eu/stats/eurofxref/eurofxref-hist.zip"

        /**
         * Load the latest exchange rate file directly from the ECB website. Use the cache by default to
         * avoid unnecessary requests.
         */
        fun fromWeb(useCache: Boolean = true) = ECBExchangeRates(DEFAULT_ECB_URL, compressed = true, useCache)

        /**
         * Load the rates from a local file
         *
         * @param path
         * @param compressed
         */
        fun fromFile(path: String, compressed: Boolean = false) =
            ECBExchangeRates("file:$path", compressed = compressed)
    }


    init {
        converters = load(url, compressed, useCache)
        logger.info {
            "loaded conversion rates from ${timeline.first()} to ${timeline.last()}"
        }
    }


    /**
     * See [CurrencyConverter.convert]
     *
     * @param from The currency to convert from
     * @param to The currency that the money needs to be converted to
     * @param amount The amount (in from currency denoted) that needs to be converted
     * @param now The time of conversion
     * @return
     */
    override fun convert(from: Currency, to: Currency, amount: Double, now: Instant): Double {
        (from == to) && return amount

        val entry = converters.floorEntry(now)?.value
        if (entry == null) {
            throw Exception("Couldn't convert currency at $now, conversion rates only start at ${converters.firstKey()}")
        } else {
            return entry.convert(from, to, amount, now)
        }
    }

    /**
     * Get the cache url of an asset. If not exist cache it now.
     *
     * @param url
     * @return
     */
    private fun cache(url: URL): URL {
        val fileName = File(url.path).name
        val day = LocalDate.now().toEpochDay()
        val cacheFile = Config.home / "$day-$fileName"

        if (cacheFile.notExists()) {
            logger.fine { "Caching $url as $cacheFile " }
            url.openStream().use {
                Files.copy(it, cacheFile, StandardCopyOption.REPLACE_EXISTING)
            }
        }
        return cacheFile.toUri().toURL()
    }

    /**
     * Load the rates from a URL, either a local file or directly from the ECB website. When
     * compressed is true, this file will be handled assumed to be in zip format and first decompressed before
     * parsed.
     *
     * The entries have only a date stamp, but according to the ECB website, they are typically published around
     * 16:00 Brussels time. So that is used to generate the final timestamps.
     *
     * @param urlString
     * @param compressed
     * @return
     */
    private fun load(urlString: String, compressed: Boolean, cache: Boolean): TreeMap<Instant, FixedExchangeRates> {
        var url = URL(urlString)

        /**
         * See if we need to cache it. We get it max once per day in that case and don't put high load on ECB website
         */
        if (cache && url.protocol.uppercase().startsWith("HTTP")) {
            url = cache(url)
        }

        var input: InputStream = url.openStream()

        // Unzip the file and use the first entry as the input stream.
        if (compressed) {
            val zis = ZipInputStream(input)
            val entry = zis.nextEntry
            input = if (entry != null) {
                logger.fine { "Found file ${entry.name} from $url" }
                zis
            } else {
                logger.severe { "File $urlString is not compressed" }
                input.close()
                url.openStream()
            }
        }
        val inputReader = InputStreamReader(input)
        val reader = CsvReader.builder().build(inputReader)

        val lines = reader.map { it.fields.toTypedArray() }
        reader.close()
        val currencies = lines.first().drop(1).filter { it.isNotBlank() }.map { Currency.getInstance(it) }

        // The timezone of the ECB
        val zoneId = ZoneId.of("Europe/Brussels")
        val dtf = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss").withZone(zoneId)

        val rates = TreeMap<Instant, FixedExchangeRates>()
        for (line in lines.drop(1).reversed()) {

            // According to the ECB website, the rates are published around 16:00 local time
            val instant = ZonedDateTime.parse(line[0] + " 16:00:00", dtf).toInstant()

            val m = mutableMapOf<Currency, Double>()
            for ((i, rateStr) in line.drop(1).withIndex()) {
                try {
                    val v = 1.0 / rateStr.toDouble()
                    val k = currencies[i]
                    m[k] = v
                } catch (ex: NumberFormatException) {
                    // Will happen due to N/A values and trailing comma in the CSV file
                    logger.fine { "Encounter number format exception for string $rateStr" }
                }
            }
            rates[instant] = FixedExchangeRates(Currency.EUR, m)
        }
        return rates
    }

}